import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from .tools import geometry, operators, sanitize


# =============================================================================
# Geostrophy
# =============================================================================

def geostrophy(
        ssh_t: Float[Array, "lat lon"],
        lat_t: Float[Array, "lat lon"],
        lon_t: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"] = None,
        return_grids: bool = True
) -> [Float[Array, "lat lon"], ...]:
    """
    Computes the geostrophic Sea Surface Current (SSC) velocity field from a Sea Surface Height (SSH) field.

    The geostrophic SSC velocity field is computed on a C-grid, following NEMO convention [1]_.

    Parameters
    ----------
    ssh_t : Float[Array, "lat lon"]
        SSH field (on the T grid)
    lat_t : Float[Array, "lat lon"]
        Latitudes of the T grid
    lon_t : Float[Array, "lat lon"]
        Longitudes of the T grid
    mask : Float[Array, "lat lon"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        If not provided, inferred from ``ssh_t`` `nan` values
    return_grids : bool, optional
        If `True`, returns the U and V grids.

        Defaults to `True`

    Returns
    -------
    u_geos_u : Float[Array, "lat lon"]
        U component of the geostrophic SSC velocity field (on the U grid)
    v_geos_v : Float[Array, "lat lon"]
        V component of the geostrophic SSC velocity field (on the V grid)
    lat_u : Float[Array, "lat lon"]
        Latitudes of the U grid, if ``return_grids=True``
    lon_u : Float[Array, "lat lon"]
        Longitudes of the U grid, if ``return_grids=True``
    lat_v : Float[Array, "lat lon"]
        Latitudes of the V grid, if ``return_grids=True``
    lon_v : Float[Array, "lat lon"]
        Longitudes of the V grid, if ``return_grids=True``
    """
    # Make sure the mask is initialized
    is_land = sanitize.init_land_mask(ssh_t, mask)

    # Compute spatial steps and Coriolis factors
    dx_t, dy_t = geometry.compute_spatial_step(lat_t, lon_t)
    coriolis_factor_t = geometry.compute_coriolis_factor(lat_t)

    # Handle spurious and masked data
    ssh_t = sanitize.sanitize_data(ssh_t, 0, is_land)

    u_geos_u, v_geos_v = _geostrophy(ssh_t, dx_t, dy_t, coriolis_factor_t, is_land)

    # Handle masked data
    u_geos_u = sanitize.sanitize_data(u_geos_u, jnp.nan, is_land)
    v_geos_v = sanitize.sanitize_data(v_geos_v, jnp.nan, is_land)

    # Compute U and V grids
    lat_u, lon_u, lat_v, lon_v = geometry.compute_uv_grids(lat_t, lon_t)

    res = (u_geos_u, v_geos_v)
    if return_grids:
        res = res + (lat_u, lon_u, lat_v, lon_v)

    return res


@jax.jit
def _geostrophy(
        ssh_t: Float[Array, "lat lon"],
        dx_t: Float[Array, "lat lon"],
        dy_t: Float[Array, "lat lon"],
        coriolis_factor_t: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"]
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    # Compute the gradient of the ssh
    ssh_dx_u = operators.derivative(ssh_t, dx_t, mask, axis=1, padding="right")  # (T(i), T(i+1)) -> U(i)
    ssh_dy_v = operators.derivative(ssh_t, dy_t, mask, axis=0, padding="right")  # (T(j), T(j+1)) -> V(j)

    # Interpolate the data
    ssh_dy_t = operators.interpolation(ssh_dy_v, mask, axis=0, padding="left")  # (V(j), V(j+1)) -> T(j+1)
    ssh_dy_u = operators.interpolation(ssh_dy_t, mask, axis=1, padding="right")  # (T(i), T(i+1)) -> U(i)

    ssh_dx_t = operators.interpolation(ssh_dx_u, mask, axis=1, padding="left")  # (U(i), U(i+1)) -> T(i+1)
    ssh_dx_v = operators.interpolation(ssh_dx_t, mask, axis=0, padding="right")  # (T(j), T(j+1)) -> V(j)

    coriolis_factor_u = operators.interpolation(
        coriolis_factor_t, mask, axis=1, padding="right"
    )  # (T(i), T(i+1)) -> U(i)
    coriolis_factor_v = operators.interpolation(
        coriolis_factor_t, mask, axis=0, padding="right"
    )  # (T(j), T(j+1)) -> V(j)

    # Computing the geostrophic velocities
    u_geos_u = - geometry.GRAVITY * ssh_dy_u / coriolis_factor_u  # U(i)
    v_geos_v = geometry.GRAVITY * ssh_dx_v / coriolis_factor_v  # V(j)

    return u_geos_u, v_geos_v



from collections.abc import Callable
from functools import partial
from typing import Literal, Union

from jax import jit, lax, value_and_grad
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Array, Bool, Float, Scalar
import optax

from .tools import geometry, kinematics, sanitize
from .geostrophy import geostrophy

#: Default maximum number of iterations for Penven and Ioannou approaches
N_IT_IT = 20
#: Default residual tolerance for Penven and Ioannou approaches
RES_EPS_IT = 0.01
#: Default size of the grid points used to compute the residual for Ioannou's approach
RES_FILTER_SIZE_IT = 3

#: Default maximum number of iterations for our variational approach
N_IT_VAR = 2000
#: Default learning rate for the gradient descent for our variational approach
LR_VAR = 0.005


# =============================================================================
# Cyclogeostrophy
# =============================================================================


def _cyclogeostrophic_loss(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        u_cyclo_u: Float[Array, "lat lon"],
        v_cyclo_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        coriolis_factor_u: Float[Array, "lat lon"],
        coriolis_factor_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"]
) -> Float[Scalar, ""]:
    u_imbalance, v_imbalance = kinematics.cyclogeostrophic_imbalance(
        u_geos_u, v_geos_v, u_cyclo_u, v_cyclo_v,
        dx_u, dx_v, dy_u, dy_v, coriolis_factor_u, coriolis_factor_v,
        mask
    )

    return jnp.nansum(u_imbalance ** 2) + jnp.nansum(v_imbalance ** 2)


def cyclogeostrophy(
        ssh_t: Float[Array, "lat lon"],
        lat_t: Float[Array, "lat lon"],
        lon_t: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"] = None,
        method: Literal["variational", "iterative"] = "variational",
        n_it: int = None,
        optim: Union[optax.GradientTransformation, str] = "sgd",
        optim_kwargs: dict = None,
        res_eps: float = RES_EPS_IT,
        use_res_filter: bool = False,
        res_filter_size: int = RES_FILTER_SIZE_IT,
        return_geos: bool = False,
        return_grids: bool = True,
        return_losses: bool = False
) -> [Float[Array, "lat lon"], ...]:
    """
    Computes the cyclogeostrophic Sea Surface Current (SSC) velocity field from a Sea Surface Height (SSH) field
    using a variational (default) or iterative method.

    The cyclogeostrophic SSC velocity field is computed on a C-grid, following NEMO convention [1]_.

    Parameters
    ----------
    ssh_t : Float[Array, "lat lon"]
        SSH field (on the T grid)
    lat_t : Float[Array, "lat lon"]
        Latitude of the T grid
    lon_t : Float[Array, "lat lon"]
        Longitude of the T grid
    mask : Float[Array, "lat lon"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        If not provided, inferred from ``ssh_t`` `nan` values
    method : Literal["variational", "iterative"], optional
        Estimation method to use.
        If ``method="variational"``, then use our variational formulation.
        If ``method="iterative"``, then use an iterative approach [2]_, [3]_.

        Defaults to `variational`
    n_it : int, optional
        Maximum number of iterations.

        Defaults to ``N_IT_VAR`` or ``N_IT_IT``, based on ``method``
    optim : Union[optax.GradientTransformation, str], optional
        Optimizer to use.
        Can be an ``optax.GradientTransformation`` optimizer, or a ``string`` referring to such an optimizer [4]_.

        Defaults to `sgd`
    optim_kwargs : dict, optional
        Optimizer arguments (such as learning rate, etc...).

        If `None`, only the learning rate is enforced to ``LR_VAR``
    res_eps : float, optional
        Residual tolerance of the iterative approach.
        When residuals are smaller, the iterative approach considers local convergence to cyclogeostrophy.

        Defaults to ``RES_EPS_IT``
    use_res_filter : bool, optional
        Use of a convolution filter for the iterative approach when computing the residuals [3]_ or not [2]_.

        Defaults to `False`
    res_filter_size : int, optional
        Size of the convolution filter of the iterative approach, when ``use_res_filter=True``.

        Defaults to ``RES_FILTER_SIZE_IT``
    return_geos : bool, optional
        If `True`, returns the geostrophic SSC velocity field in addition to the cyclogeostrophic one.

        Defaults to `False`
    return_grids : bool, optional
        If `True`, returns the U and V grids.

        Defaults to `True`
    return_losses : bool, optional
        If `True`, returns the losses (cyclogeostrophic imbalance) over iterations.

        Defaults to `False`

    Returns
    -------
    u_cyclo_u : Float[Array, "lat lon"]
        U component of the cyclogeostrophic SSC velocity field (on the U grid)
    v_cyclo_v : Float[Array, "lat lon"]
        V component of the cyclogeostrophic SSC velocity field (on the V grid)
    u_geos_u : Float[Array, "lat lon"]
        U component of the geostrophic SSC velocity field (on the U grid), if ``return_geos=True``
    v_geos_v : Float[Array, "lat lon"]
        V component of the geostrophic SSC velocity field (on the V grid), if ``return_geos=True``
    lat_u : Float[Array, "lat lon"]
        Latitudes of the U grid, if ``return_grids=True``
    lon_u : Float[Array, "lat lon"]
        Longitudes of the U grid, if ``return_grids=True``
    lat_v : Float[Array, "lat lon"]
        Latitudes of the V grid, if ``return_grids=True``
    lon_v : Float[Array, "lat lon"]
        Longitudes of the V grid, if ``return_grids=True``
    losses: Float[Array, "n_it"]
        Cyclogeostrophic imbalance evaluated at each iteration, if ``return_losses=True``
    """
    # Make sure the mask is initialized
    is_land = sanitize.init_land_mask(ssh_t, mask)

    # Compute geostrophic SSC velocity field
    u_geos_u, v_geos_v, lat_u, lon_u, lat_v, lon_v = geostrophy(ssh_t, lat_t, lon_t, is_land, return_grids=True)

    # Compute spatial steps and Coriolis factors
    dx_u, dy_u = geometry.compute_spatial_step(lat_u, lon_u)
    dx_v, dy_v = geometry.compute_spatial_step(lat_v, lon_v)
    coriolis_factor_u = geometry.compute_coriolis_factor(lat_u)
    coriolis_factor_v = geometry.compute_coriolis_factor(lat_v)

    # Handle spurious and masked data
    u_geos_u = sanitize.sanitize_data(u_geos_u, 0, is_land)
    v_geos_v = sanitize.sanitize_data(v_geos_v, 0, is_land)

    if method == "variational":
        if n_it is None:
            n_it = N_IT_VAR
        if isinstance(optim, str):
            if optim_kwargs is None:
                optim_kwargs = {"learning_rate": LR_VAR}
            optim = getattr(optax, optim)(**optim_kwargs)
        elif not isinstance(optim, optax.GradientTransformation):
            raise TypeError("optim should be an optax.GradientTransformation optimizer, or a string referring to such "
                            "an optimizer.")
        res = _variational(u_geos_u, v_geos_v, dx_u, dx_v, dy_u, dy_v, coriolis_factor_u, coriolis_factor_v, is_land,
                           n_it, optim)
    elif method == "iterative":
        if n_it is None:
            n_it = N_IT_IT
        res = _iterative(u_geos_u, v_geos_v, dx_u, dx_v, dy_u, dy_v, coriolis_factor_u, coriolis_factor_v, is_land,
                         n_it, res_eps, use_res_filter, res_filter_size, return_losses)
    else:
        raise ValueError("method should be one of [\"variational\", \"iterative\"]")

    # Handle masked data
    u_cyclo_u, v_cyclo_v, losses = res
    u_cyclo_u = sanitize.sanitize_data(u_cyclo_u, jnp.nan, is_land)
    v_cyclo_v = sanitize.sanitize_data(v_cyclo_v, jnp.nan, is_land)

    res = (u_cyclo_u, v_cyclo_v)
    if return_geos:
        u_geos_u = sanitize.sanitize_data(u_geos_u, jnp.nan, is_land)
        v_geos_v = sanitize.sanitize_data(v_geos_v, jnp.nan, is_land)
        res = res + (u_geos_u, v_geos_v)
    if return_grids:
        res = res + (lat_u, lon_u, lat_v, lon_v)
    if return_losses:
        res = res + (losses,)

    return res


# =============================================================================
# Iterative method
# =============================================================================


def _it_step(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        coriolis_factor_u: Float[Array, "lat lon"],
        coriolis_factor_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        res_eps: float,
        res_filter: Float[Array, "lat lon"],
        res_weights: Float[Array, "lat lon"],
        use_res_filter: bool,
        return_losses: bool,
        u_n: Float[Array, "lat lon"],
        v_n: Float[Array, "lat lon"],
        mask_update: Bool[Array, "lat lon"],
        res_n: Float[Array, "lat lon"]
) -> [[Float[Array, "lat lon"], Float[Array, "lat lon"], Float[Array, "lat lon"], Float[Array, "lat lon"], int], float]:
    # compute loss
    loss = lax.cond(
        return_losses,
        lambda: _cyclogeostrophic_loss(
            u_geos_u, v_geos_v, u_n, v_n, dx_u, dx_v, dy_u, dy_v, coriolis_factor_u, coriolis_factor_v, mask
        ),
        lambda: jnp.nan
    )

    # next it
    u_adv_v, v_adv_u = kinematics.advection(u_n, v_n, dx_u, dx_v, dy_u, dy_v, mask)
    u_np1 = u_geos_u - jnp.nan_to_num(v_adv_u / coriolis_factor_u, copy=False, nan=0, posinf=0, neginf=0)
    v_np1 = v_geos_v + jnp.nan_to_num(u_adv_v / coriolis_factor_v, copy=False, nan=0, posinf=0, neginf=0)

    # compute dist to u_cyclo and v_cyclo
    res_np1 = jnp.abs(u_np1 - u_n) + jnp.abs(v_np1 - v_n)  # norm1
    res_np1 = lax.cond(
        use_res_filter,  # apply filter
        lambda operands: jsp.signal.convolve(operands[0], operands[1], mode="same", method="fft") / operands[2],
        lambda operands: operands[0],
        (res_np1, res_filter, res_weights)
    )

    # compute stopping criterion masks
    mask_not_div = jnp.where(res_np1 <= res_n, True, False)
    mask_not_conv = jnp.where(res_np1 >= res_eps, True, False)
  
    # update cyclogeostrophic velocities and residuals where it is not diverging
    mask_update &= mask_not_div
    u_n = jnp.where(mask_update, u_np1, u_n)
    v_n = jnp.where(mask_update, v_np1, v_n)
    res_n = jnp.where(mask_update, res_np1, res_n)

    # update stopping criterion mask where it has converged
    mask_update &= mask_not_conv

    return (u_n, v_n, mask_update, res_n), loss


@partial(jit, static_argnames=("n_it", "res_filter_size"))
def _iterative(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        coriolis_factor_u: Float[Array, "lat lon"],
        coriolis_factor_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        n_it: int,
        res_eps: float,
        use_res_filter: bool,
        res_filter_size: int,
        return_losses: bool
) -> [Float[Array, "lat lon"], ...]:
    # used if applying a filter when computing stopping criteria
    res_filter = jnp.ones((res_filter_size, res_filter_size))
    res_weights = jsp.signal.convolve(jnp.ones_like(u_geos_u), res_filter, mode="same", method="fft")

    # define step partial: freeze constant over iterations
    def step_fn(carry, _):
        return _it_step(
            u_geos_u, v_geos_v,
            dx_u, dx_v, dy_u, dy_v,
            coriolis_factor_u, coriolis_factor_v, mask,
            res_eps, res_filter, res_weights,
            use_res_filter, return_losses,
            *carry
        )

    # apply updates
    (u_cyclo, v_cyclo, _, _), losses = lax.scan(
        step_fn,
        (u_geos_u, v_geos_v, (1 - mask).astype(bool), jnp.maximum(jnp.abs(u_geos_u), jnp.abs(v_geos_v))),
        xs=None, length=n_it
    )

    return u_cyclo, v_cyclo, losses


# =============================================================================
# Variational method
# =============================================================================

def _var_loss_fn(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        coriolis_factor_u: Float[Array, "lat lon"],
        coriolis_factor_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        uv_cyclo: [Float[Array, "lat lon"], Float[Array, "lat lon"]]
) -> Float[Scalar, ""]:
    u_cyclo_u, v_cyclo_v = uv_cyclo
    return _cyclogeostrophic_loss(
        u_geos_u, v_geos_v, u_cyclo_u, v_cyclo_v, dx_u, dx_v, dy_u, dy_v,  coriolis_factor_u, coriolis_factor_v, mask
    )


def _var_step(
        loss_fn: Callable[[[Float[Array, "lat lon"], Float[Array, "lat lon"]]], Float[Scalar, ""]],
        optim: optax.GradientTransformation,
        u_cyclo_u: Float[Array, "lat lon"],
        v_cyclo_v: Float[Array, "lat lon"],
        opt_state: optax.OptState
) -> [[Float[Array, "lat lon"], ...], float]:
    params = (u_cyclo_u, v_cyclo_v)
    # evaluate the cost function and compute its gradient
    loss, (u_grad, v_grad) = value_and_grad(loss_fn)(params)
    u_grad = jnp.nan_to_num(u_grad, copy=False, nan=0, posinf=0, neginf=0)
    v_grad = jnp.nan_to_num(v_grad, copy=False, nan=0, posinf=0, neginf=0)
    # update the optimizer
    updates, opt_state = optim.update((u_grad, v_grad), opt_state, params)
    # apply updates to the parameters
    u_n, v_n = optax.apply_updates(params, updates)

    return (u_n, v_n, opt_state), loss


def _solve(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        loss_fn: Callable[[[Float[Array, "lat lon"], Float[Array, "lat lon"]]], Float[Scalar, ""]],
        n_it: int,
        optim: optax.GradientTransformation
) -> [Float[Array, "lat lon"], ...]:
    # define step partial: freeze constant over iterations
    def step_fn(carry, _):
        return _var_step(loss_fn, optim, *carry)

    (u_cyclo_u, v_cyclo_v, _), losses = lax.scan(
        step_fn,
        (u_geos_u, v_geos_v, optim.init((u_geos_u, v_geos_v))),
        xs=None, length=n_it
    )

    return u_cyclo_u, v_cyclo_v, losses


@partial(jit, static_argnames=("n_it", "optim"))
def _variational(
        u_geos_u: Float[Array, "lat lon"],
        v_geos_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        coriolis_factor_u: Float[Array, "lat lon"],
        coriolis_factor_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        n_it: int,
        optim: optax.GradientTransformation
) -> [Float[Array, "lat lon"], ...]:
    # define loss partial: freeze constant over iterations
    loss_fn = partial(
        _var_loss_fn,
        u_geos_u, v_geos_v, dx_u, dx_v, dy_u, dy_v, coriolis_factor_u, coriolis_factor_v, mask
    )

    return _solve(u_geos_u, v_geos_v, loss_fn, n_it, optim)



from jax import lax
import jax.numpy as jnp
from jaxtyping import Array, Float
import numpy as np
from scipy import interpolate


def sanitize_data(
        arr: Float[Array, "lat lon"],
        fill_value: float,
        mask: Float[Array, "lat lon"]
) -> Float[Array, "lat lon"]:
    """
    Sanitizes data by replacing `nan` with ``fill_value`` and applying ``fill_value`` to the masked area.

    Parameters
    ----------
    arr : Float[Array, "lat lon"]
        Array to sanitize
    fill_value : float
        Value to replace `nan` values and masked area with
    mask :  Float[Array, "lat lon"]
        Mask to apply, `1` or `True` for masked

    Returns
    -------
    arr : Float[Array, "lat lon"]
        Sanitized array
    """
    arr = jnp.nan_to_num(arr, copy=False, nan=fill_value, posinf=fill_value, neginf=fill_value)
    arr = jnp.where(mask, fill_value, arr)
    return arr


def init_land_mask(
        field: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"] = None
) -> Float[Array, "lat lon"]:
    """
    If ``mask is None``, initializes it from the `nan` values of ``field``.
    If ``mask is not None``, simply returns it.

    Parameters
    ----------
    field : Float[Array, "lat lon"]
        Field used to initialize the mask (if needed)
    mask :  Float[Array, "lat lon"], optional
        Mask to initialized (if `None`).

        Defaults to `None`

    Returns
    -------
    mask : Float[Array, "lat lon"]
        Initialized (if needed) mask
    """
    if mask is None:
        mask = ~jnp.isfinite(field)
    return mask


def handle_land_boundary(
        field1: Float[Array, "lat lon"],
        field2: Float[Array, "lat lon"],
        mask1: Float[Array, "lat lon"],
        mask2: Float[Array, "lat lon"],
        pad_left: bool
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    """
    Replaces the masked values of ``field1`` (``field2``) with values of ``field2`` (``field1``), element-wise.

    It allows computing more coherent values when applying grid operators.
    In such cases, ``field1`` and ``field2`` are left and right shifted versions of a field (along one of the axes).

    Parameters
    ----------
    field1 : Float[Array, "lat lon"]
        A field
    field2 : Float[Array, "lat lon"]
        Another field
    mask1 : Float[Array, "lat lon"]
        A mask defining the marine area of ``field1`` spatial domain; `1` or `True` stands for masked (i.e. land)
    mask2 : Float[Array, "lat lon"]
        A mask defining the marine area of ``field2`` spatial domain; `1` or `True` stands for masked (i.e. land)
    pad_left : bool
        If `True`, apply padding in the `left` direction (i.e. `West` or `South`) ;
        if `False`, apply padding in the `right` direction (i.e. `East` or `North`).

    Returns
    -------
    field1 : Float[Array, "lat lon"]
        A field whose masked values have been replaced with the ones from ``field2``
    field2 : Float[Array, "lat lon"]
        A field whose masked values have been replaced with the ones from ``field1``
    """
    field1, field2 = lax.cond(
        pad_left,
        lambda: (jnp.where(mask1, field2, field1), field2),
        lambda: (field1, jnp.where(mask2, field1, field2))
    )
    return field1, field2


def sanitize_grid_np(
        lat: Float[Array, "lat lon"],
        lon: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"] = None
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    """
    Sanitizes (unstructured) grids by interpolated and extrapolated `nan` or masked values to avoid spurious
    (`0`, `nan`, `inf`) spatial steps and Coriolis factors.

    Helper function written using pure ``numpy`` and ``scipy``, and as such not used internally,
    because incompatible with ``jax.vmap`` and likes.
    Should be used before calling ``jaxparrow.geostrophy`` or ``jaxparrow.cyclogeostrophy``
    in case of suspicious latitudes or longitudes T grids.

    Caution: because it uses ``scipy.interpolate.RBFInterpolator``,
    it's memory usage grows quadratically with the number of grid points.

    Parameters
    ----------
    lat : Float[Array, "lat lon"]
        Grid latitudes
    lon : Float[Array, "lat lon"]
        Grid longitudes
    mask :  Float[Array, "lat lon"], optional
        Mask to apply, `1` or `True` for masked, defaults to `None`

    Returns
    -------
    lat : Float[Array, "lat lon"]
        Grid latitudes
    lon : Float[Array, "lat lon"]
        Grid longitudes
    """
    def fill_nan(arr: Float[Array, "lat lon"]) -> Float[Array, "lat lon"]:
        x = np.arange(0, arr.shape[1])
        y = np.arange(0, arr.shape[0])
        # mask invalid values
        arr = np.ma.masked_invalid(arr)
        xx, yy = np.meshgrid(x, y)
        # get only the valid values
        valid_x = xx[~arr.mask]
        valid_y = yy[~arr.mask]
        valid_arr = arr[~arr.mask]
        rbf = interpolate.RBFInterpolator(np.array([valid_x, valid_y]).T, valid_arr)
        # get the invalid ones
        invalid_x = xx[arr.mask]
        invalid_y = yy[arr.mask]
        invalid_arr = rbf(np.array([invalid_x, invalid_y]).T)
        # fill
        arr[arr.mask] = invalid_arr
        return arr.data

    # make sure nan are used behind masked pixels (and not 0)
    lat = sanitize_data(lat, jnp.nan, mask)
    lon = sanitize_data(lon, jnp.nan, mask)
    # fill nan using RBF interpolation
    lat = fill_nan(lat)
    lon = fill_nan(lon)
    return lat, lon



import jax.numpy as jnp
from jaxtyping import Array, Float

from .operators import interpolation


#: Approximate earth angular speed
EARTH_ANG_SPEED = 7.292115e-5
#: Approximate earth radius
EARTH_RADIUS = 6370e3
#: Approximate gravity
GRAVITY = 9.81
#: Degrees / radians conversion factor
P0 = jnp.pi / 180


def compute_spatial_step(
        lat: Float[Array, "lat lon"],
        lon: Float[Array, "lat lon"]
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    """
    Computes the spatial steps of a grid (in meters) along `x` and `y`.

    It makes use of the distance-on-a-sphere formula with Taylor expansion approximations of `cos` and `arccos`
    functions to avoid truncation issues.

    Parameters
    ----------
    lat : Float[Array, "lat lon"]
        Latitude grid
    lon : Float[Array, "lat lon"]
        Longitude grid

    Returns
    -------
    dx : Float[Array, "lat lon"]
        Spatial steps in meters along `x`
    dy : Float[Array, "lat lon"]
        Spatial steps in meters along `y`
    """
    def sphere_distance(_lats, _late, _lons, _lone):
        dlat, dlon = P0 * (_late - _lats), P0 * (_lone - _lons)
        return EARTH_RADIUS * jnp.sqrt(dlat ** 2 + jnp.cos(P0 * _lats) * jnp.cos(P0 * _late) * dlon ** 2)

    dx, dy = jnp.zeros_like(lat), jnp.zeros_like(lat)
    # dx
    dx = dx.at[:, :-1].set(sphere_distance(lat[:, :-1], lat[:, 1:], lon[:, :-1], lon[:, 1:]))
    dx = dx.at[:, -1].set(dx[:, -2])
    # dy
    dy = dy.at[:-1, :].set(sphere_distance(lat[:-1, :], lat[1:, :], lon[:-1, :], lon[1:, :]))
    dy = dy.at[-1, :].set(dy[-2, :])
    return dx, dy


def compute_coriolis_factor(
        lat: Float[Array, "lat lon"]
) -> Float[Array, "lat lon"]:
    """
    Computes the Coriolis factor from a latitude grid.

    Parameters
    ----------
    lat : Float[Array, "lat lon"]
        Latitudes grid

    Returns
    -------
    cf : Float[Array, "lat lon"]
        Coriolis factor grid
    """
    return 2 * EARTH_ANG_SPEED * jnp.sin(lat * P0)


def compute_uv_grids(
        lat_t: Float[Array, "lat lon"],
        lon_t: Float[Array, "lat lon"]
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"], Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    """
    Computes the U and V grids associated to a T grid following NEMO convention [1]_.

    Parameters
    ----------
    lat_t : Float[Array, "lat lon"]
        Latitudes of the T grid
    lon_t : Float[Array, "lat lon"]
        Longitudes of the T grid

    Returns
    -------
    lat_u : Float[Array, "lat lon"]
        Latitudes of the U grid
    lon_u : Float[Array, "lat lon"]
        Longitudes of the U grid
    lat_v : Float[Array, "lat lon"]
        Latitudes of the V grid
    lon_v : Float[Array, "lat lon"]
        Longitudes of the V grid
    """
    lat_mask = jnp.zeros_like(lat_t, dtype=bool)
    lon_mask = jnp.zeros_like(lon_t, dtype=bool)

    lat_u = interpolation(lat_t, lat_mask, axis=1, padding="right")
    lat_u = lat_u.at[:, -1].set(2 * lat_t[:, -1] - lat_t[:, -2])
    lon_u = interpolation(lon_t, lon_mask, axis=1, padding="right")
    lon_u = lon_u.at[:, -1].set(2 * lon_t[:, -1] - lon_t[:, -2])

    lat_v = interpolation(lat_t, lat_mask, axis=0, padding="right")
    lat_v = lat_v.at[-1, :].set(2 * lat_t[-1, :] - lat_t[-2, :])
    lon_v = interpolation(lon_t, lon_mask, axis=0, padding="right")
    lon_v = lon_v.at[-1, :].set(2 * lon_t[-1, :] - lon_t[-2, :])

    return lat_u, lon_u, lat_v, lon_v



from typing import Literal

from jax import lax
import jax.numpy as jnp
from jaxtyping import Array, Float

from .sanitize import handle_land_boundary


def interpolation(
        field: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        axis: Literal[0, 1],
        padding: Literal["left", "right"]
) -> Float[Array, "lat lon"]:
    """
    Interpolates the values of a ``field`` along a given ``axis`` (`0` for `lat`/`y`, `1` for `lon`/`x`),
    applying ``padding`` to the `left` (i.e. `West` if ``axis=1``, `South` if ``axis=0``) or
    to the `right` (i.e. `East` if ``axis=1``, `North` if ``axis=0``) of the domain,
    using nearest non interpolated value at the padded edge.

    Parameters
    ----------
    field : Float[Array, "lat lon"]
        Field to interpolate
    mask : Float[Array, "lat lon"]
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    axis : Literal[0, 1]
        Axis along which interpolation is performed
    padding : Literal["left", "right"]
        Padding direction.
        For example, following NEMO convention [1]_,
        interpolating from U to T points requires a `left` padding
        (the midpoint between $U_i$ and $U_{i+1}$ corresponds to $T_{i+1}$),
        and interpolating from T to U points a `right` padding
        (the midpoint between $T_i$ and $T_{i+1}$ corresponds to $U_i$)

    Returns
    -------
    field : Float[Array, "lat lon"]
        Interpolated field
    """
    def do_interpolate(field_b, field_f, mask_b, mask_f, pad_left):
        field_b, field_f = handle_land_boundary(field_b, field_f, mask_b, mask_f, pad_left)
        return 0.5 * (field_b + field_f)

    def axis0(pad_left):
        field_b, field_f = field[:-1, :], field[1:, :]
        mask_b, mask_f = mask[:-1, :], mask[1:, :]
        midpoint_values = do_interpolate(field_b, field_f, mask_b, mask_f, pad_left)

        arr = lax.cond(
            pad_left,
            lambda: jnp.pad(midpoint_values, pad_width=((1, 0), (0, 0)), mode="edge"),
            lambda: jnp.pad(midpoint_values, pad_width=((0, 1), (0, 0)), mode="edge")
        )

        return arr

    def axis1(pad_left):
        field_b, field_f = field[:, :-1], field[:, 1:]
        mask_b, mask_f = mask[:, :-1], mask[:, 1:]
        midpoint_values = do_interpolate(field_b, field_f, mask_b, mask_f, pad_left)

        arr = lax.cond(
            pad_left,
            lambda: jnp.pad(midpoint_values, pad_width=((0, 0), (1, 0)), mode="edge"),
            lambda: jnp.pad(midpoint_values, pad_width=((0, 0), (0, 1)), mode="edge")
        )

        return arr

    field = lax.cond(
        axis == 0,
        lambda pad_left: axis0(pad_left),
        lambda pad_left: axis1(pad_left),
        padding == "left"
    )

    return field


def derivative(
        field: Float[Array, "lat lon"],
        dxy: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"],
        axis: Literal[0, 1],
        padding: Literal["left", "right"]
) -> Float[Array, "lat lon"]:
    """
    Differentiates a ``field``, using finite differences, along a given ``axis`` (`0` for `lat`/`y`, `1` for `lon`/'x'),
    applying ``padding`` to the `left` (i.e. `West` if ``axis=1``, `South` if ``axis=0``) or
    to the `right` (i.e. `East` if ``axis=1``, `North` if ``axis=0``) of the domain,
    using nearest derivative value at the padded edge.

    Parameters
    ----------
    field : Float[Array, "lat lon"]
        Field to differentiate
    dxy : Float[Array, "lat lon"]
        Spatial steps
    mask : Float[Array, "lat lon"]
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    axis : Literal[0, 1]
        Axis along which interpolation is performed
    padding : Literal["left", "right"]
        Padding direction.
        For example, following NEMO convention [1]_,
        interpolating from U to T points requires a `left` padding
        (the midpoint between $U_i$ and $U_{i+1}$ corresponds to $T_{i+1}$),
        and interpolating from T to U points a `right` padding
        (the midpoint between $T_i$ and $T_{i+1}$ corresponds to $U_i$)

    Returns
    -------
    field : Float[Array, "lat lon"]
        Interpolated field
    """
    def do_differentiate(field_b, field_f, mask_b, mask_f, pad_left):
        field_b, field_f = handle_land_boundary(field_b, field_f, mask_b, mask_f, pad_left)
        return field_f - field_b

    def axis0(pad_left):
        field_b, field_f = field[:-1, :], field[1:, :]
        mask_b, mask_f = mask[:-1, :], mask[1:, :]
        midpoint_values = do_differentiate(field_b, field_f, mask_b, mask_f, pad_left)

        arr = lax.cond(
            pad_left,
            lambda: jnp.pad(midpoint_values, pad_width=((1, 0), (0, 0)), mode="edge"),
            lambda: jnp.pad(midpoint_values, pad_width=((0, 1), (0, 0)), mode="edge")
        )

        return arr

    def axis1(pad_left):
        field_b, field_f = field[:, :-1], field[:, 1:]
        mask_b, mask_f = mask[:, :-1], mask[:, 1:]
        midpoint_values = do_differentiate(field_b, field_f, mask_b, mask_f, pad_left)

        arr = lax.cond(
            pad_left,
            lambda: jnp.pad(midpoint_values, pad_width=((0, 0), (1, 0)), mode="edge"),
            lambda: jnp.pad(midpoint_values, pad_width=((0, 0), (0, 1)), mode="edge")
        )

        return arr

    field = lax.cond(
        axis == 0,
        lambda pad_left: axis0(pad_left),
        lambda pad_left: axis1(pad_left),
        padding == "left"
    )

    return field / dxy



import jax
import jax.numpy as jnp

from jaxparrow.tools import geometry, kinematics


def simulate_gaussian_eddy(
        r0: float,
        dxy: float,
        eta0: float,
        latitude: int
) -> [jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array,
      jax.Array]:
    l0 = r0 * 2  # limit boundary impact
    xy = jnp.arange(0, l0, dxy)
    xy = jnp.concatenate((-xy[::-1][:-1], xy))
    X, Y = jnp.meshgrid(xy, xy)
    R = jnp.hypot(X, Y)
    dXY = jnp.ones_like(X) * dxy
    coriolis_factor = jnp.ones_like(R) * geometry.compute_coriolis_factor(latitude)  # noqa

    ssh = simulate_gaussian_ssh(r0, eta0, R)
    u_geos_t, v_geos_t = simulate_gaussian_geos(r0, X, Y, ssh, coriolis_factor)
    u_cyclo_t, v_cyclo_t = simulate_gaussian_cyclo(r0, jnp.arctan2(Y, X), u_geos_t, v_geos_t, coriolis_factor)

    mask = jax.numpy.full_like(ssh, 0, dtype=bool)  # no land in square oceans

    return X, Y, R, dXY, coriolis_factor, ssh, u_geos_t, v_geos_t, u_cyclo_t, v_cyclo_t, mask


def simulate_gaussian_ssh(r0: float, eta0: float, R: jax.Array) -> jax.Array:
    return eta0 * jnp.exp(-(R / r0)**2)


def simulate_gaussian_geos(
        r0: float,
        X: jax.Array,
        Y: jax.Array,
        ssh: jax.Array,
        coriolis_factor: jax.Array
) -> [jax.Array, jax.Array]:
    def f():
        return 2 * geometry.GRAVITY * ssh / (coriolis_factor * r0 ** 2)
    u_geos = Y * f()
    v_geos = -X * f()
    return u_geos, v_geos


def simulate_gaussian_cyclo(
        r0: float,
        theta: jax.Array,
        u_geos: jax.Array,
        v_geos: jax.Array,
        coriolis_factor: jax.Array
) -> [jax.Array, jax.Array]:
    def f():
        return azim_cyclo**2 / (r0 * coriolis_factor)
    azim_geos = kinematics.magnitude(u_geos, v_geos, interpolate=False)
    azim_cyclo = 2 * azim_geos / (1 + jnp.sqrt(1 + 4 * azim_geos / (coriolis_factor * r0)))
    u_cyclo = u_geos + jnp.sin(theta) * f()
    v_cyclo = v_geos - jnp.cos(theta) * f()
    return u_cyclo, v_cyclo


def compute_rmse(y: jax.Array, y_hat: jax.Array) -> jax.Array:
    return jnp.sqrt(jnp.nanmean((y - y_hat)**2))
